module Handlers
 (
  emergencyHandler,
  attackHandler,
  checkHandler,
  pickHandler
 ) where

import FiniteMap
import Set
import List
import Maybe
import IArray
import IOExts

import Data
import Utils
import Bfs
import Checks

checkDangerousPosition::Board->Self->State->Maybe [Dir]
checkDangerousPosition board self state
 | isEmptyFM dangerousRobots = Nothing
 | otherwise = Just dDirs
 where
  robots = sRobots state
  me = findMe state self
  myPos@(mX,mY) = rPos me
  closeRobots = filterFM (\_ r -> vic myPos (rPos r) == 1) robots
  dangerousRobots = filterFM (\_ -> isDangerous) closeRobots
  dPoints = map head $ group $ sort $ map rPos $ eltsFM dangerousRobots
  dDirs = map pointToDir $ map (\p->p `subPoint` myPos) dPoints
  isDangerous r = bC board ! iPos == FWater
   where
    (hX,hY) = rPos r
    iPos = (2*mX - hX, 2*mY - hY)

genEscapeMove::[Dir]->Board->Self->State->Maybe [Move]
genEscapeMove dDirs board self state
 | null aMoves = Just [Move prize (MMove $ head dDirs)]
 | otherwise = Just [Move prize (MMove $ snd $ head aMoves)]
 where
  me = findMe state self
  prize = 1 `max` (sMoney self `div` 16)
  moves = map (\d->(rPos me `move` d,d)) dirList
  pMoves = filter (\(pos,_)->bC board ! pos == FEmpty) moves
  aMoves = filter mayPass pMoves
  mayPass (pos,dir)
   | bC board ! pos /= FEmpty = False
   | isEmptyFM (locateRobot state pos) = True
   | otherwise = mayPass (pos `move` dir, dir)

escapeMoveStrategy::[Dir]->Strategy
escapeMoveStrategy dDirs =
 Strategy
  {
   strName = "Escape attacks from " ++ show dDirs,
   strGenerate = genEscapeMove dDirs,
   strCheck = \_ _ _ _ -> alwaysOK,
   strHandlers = []
  }

emergencyHandler::Handler
emergencyHandler = Handler checkDangerousPosition escapeMoveStrategy False

checkTarget::Bool->Int->Board->Self->State->Maybe Int
checkTarget rob range board self state
  | not (null toKill) = Just chToKill
  | rob && not (null toRob) = Just chToRob
  | otherwise = Nothing
 where
  me = findMe state self
  pos = rPos me
  robots = sRobots state
  closeRobots' = mapMaybe isCloseRobot $ fmToList robots
  closeRobots = filter (\(d,_,_) -> d>0) closeRobots'
  adjacentRobots = filter (\(d,_,_) -> d==1) closeRobots
  toKill = filter killable adjacentRobots
  (_,chToKill,_) = minimumBy (\(d1,_,_) (d2,_,_)->compare d1 d2) toKill
  toRob = filter robable closeRobots
  (_,chToRob,_) = minimumBy (\(d1,_,_) (d2,_,_)->compare d1 d2) toRob

  robable (_,_,r) = not (null iPkgs)
   where
    packages = rPackages r
    pkgs = map (findPackage state) $ setToList packages
    iPkgs = filter (testWeight self) $ filter (testDestination board) pkgs
  killable (_,_,r) = bC board ! d == FWater
   where
    d = rPos r `addPoint` rPos r `subPoint` pos
  cDists = closeDistances board pos range
  isCloseRobot (i,r) = fmap (\p->(length p,i,r)) $ closePath cDists pos (rPos r) range

genAttackMove::Bool->Int->Int->Board->Self->State->Maybe [Move]
genAttackMove rob range i board self state
 | Nothing <- robot = Nothing
 | Nothing <- way = Nothing
 | not rob && not killable = Nothing
 | not robable && not killable = Nothing
 | length way' == 1 = Just [Move prize (MMove $ head way')]
 | otherwise = Just [Move 1 (MMove $ head way')]
 where
  prize = 2 `max` (sMoney self `div` 16)
  robots = sRobots state
  robot = lookupFM robots i
  Just r = robot
  me = findMe state self
  myPos = rPos me
  hisPos = rPos r
  killable = vic myPos hisPos == 1 && bC board ! d == FWater
   where
    d = hisPos `addPoint` hisPos `subPoint` myPos
  robable = not (null iPkgs)
   where
    packages = rPackages r
    pkgs = map (findPackage state) $ setToList packages
    iPkgs = filter (testWeight self) $ filter (testDestination board) pkgs
  cDists = closeDistances board myPos range
  way = closePath cDists myPos hisPos range
  Just way' = way

attackTargetStrategy::Bool->Int->Int->Strategy
attackTargetStrategy rob range i =
  Strategy {strName = "Attack " ++ show i ++ (if rob then " (rob)" else ""),
            strGenerate = genAttackMove rob range i,
            strCheck = genCheck,
            strHandlers = []}
 where
  genCheck board self state (Move 1 _) = alwaysWrong
  genCheck board self state _ = alwaysOK

attackHandler::Bool->Int->Handler
attackHandler rob range = Handler (checkTarget rob range) (attackTargetStrategy rob range) rob

checkUnknownPackage::Int->Board->Self->State->Maybe (Either (Int,Package) Point)
checkUnknownPackage range board self state
 | null closePack && null closeBase = Nothing
 | null closeBase = Just $ Left (i,findPackage state i)
 | otherwise = Just $ Right basePos
 where
  me = findMe state self
  packages = sPackages state
  bases = sBases state
  unidPackages = filter unidPack $ fmToList packages
  unidPack (_,pack) = isLeft (pPos pack) && (not (pSure pack) || isNothing (pDest pack))
  closePack = mapMaybe isClosePack unidPackages
  closeBase = mapMaybe isCloseBase $ setToList bases
  (_,i) = minimum closePack
  (_,basePos) = minimum closeBase
  myPos = rPos me
  cDists = closeDistances board myPos range
  isClosePack (i,pack) = fmap (\p->(length p,i)) $ closePath cDists myPos ppos range
   where
    Left ppos = pPos pack
  isCloseBase pos = fmap (\p->(length p,pos)) $ closePath cDists myPos pos range

genCloseMoveToMoves::Check->Int->Point->Board->Self->State->Maybe [Move]
genCloseMoveToMoves check range to board self state
  | check board self state =
      fmap (map (\d -> Move 1 (MMove d))) $ closePath cDists myPos to range
  | otherwise = Nothing
 where
  me = findMe state self
  myPos = rPos me
  cDists = closeDistances board myPos range

identifyPackageStrategy::Int->Either (Int,Package) Point->Strategy
identifyPackageStrategy range (Left (i,pack)) =
 Strategy {strName = "Identify packages at " ++ show pos,
           strGenerate = genCloseMoveToMoves (checkPackageAt i pos) range pos,
           strCheck = genCheck,
           strHandlers = [attackHandler False 1, emergencyHandler]}
 where
  Left pos = pPos pack
  genCheck board self state (Move _ (MMove dir)) =
    checkPackageAt i pos `andCheck`
    checkAt (rPos (findMe state self) `move` dir)
identifyPackageStrategy range (Right base) =
 Strategy {strName = "Identify base at " ++ show base,
           strGenerate = genCloseMoveToMoves (checkBaseExists base) range base,
           strCheck = genCheck,
           strHandlers = [attackHandler False 1, emergencyHandler]}
 where
  genCheck board self state (Move _ (MMove dir)) =
    checkAt (rPos (findMe state self) `move` dir) `andCheck`
    checkBaseExists base

checkHandler::Int->Handler
checkHandler range = Handler (checkUnknownPackage range) (identifyPackageStrategy range) True

checkPickablePackage::Int->Board->Self->State->Maybe (Int,Package)
checkPickablePackage range board self state
 | not (isEmptyFM closeRobots) = Nothing
 | null closePack = Nothing
 | otherwise = Just (i,findPackage state i)
 where
  me = findMe state self
  packages = sPackages state
  bases = sBases state
  idPackages = filter mayCarry $ filter idPack $ fmToList packages
  idPack (_,pack) = isLeft (pPos pack) && pSure pack && not (isNothing $ pDest pack)
  myCapa = capa state self
  mayCarry (i,pack) = testDestination board pack &&
                      fromJust (pWeight pack) <= myCapa
 
  closePack = mapMaybe isClosePack idPackages
  (_,i) = minimum closePack
  myPos = rPos me

  robots = sRobots state
  closeRobots = filterFM (\_ r -> let d = vic myPos (rPos r) in d > 0 && d <= 2) robots

  cDists = closeDistances board myPos range
  isClosePack (i,pack) = fmap (\p->(length p,i)) $ closePath cDists myPos ppos range
   where
    Left ppos = pPos pack

getToPackageStrategy::Int->(Int,Package)->Strategy
getToPackageStrategy range (i,pack) =
 Strategy {strName = "Get to packages at " ++ show pos,
           strGenerate = genCloseMoveToMoves (checkPackageAt i pos) range pos,
           strCheck = genCheck,
           strHandlers = [attackHandler False 1, emergencyHandler]}
 where
  Left pos = pPos pack
  genCheck board self state (Move _ (MMove dir)) =
    checkPackageAt i pos `andCheck`
    checkAt (rPos (findMe state self) `move` dir)

pickHandler::Int->Handler
pickHandler range = Handler (checkPickablePackage range) (getToPackageStrategy range) True

