module Strategies
 (
  defaultStrategy
 ) where

import FiniteMap
import Set
import IArray
import List
import Maybe

import Data
import Utils
import Checks
import Bfs
import Handlers

genMoveToBaseMoves::Point->UArray Point Int->Board->Self->State->Maybe [Move]
genMoveToBaseMoves to dists board self state
  | not (checkBaseExists to board self state) = Nothing
  | otherwise = Just $ map (\d -> Move 1 (MMove d)) $ rPath dists (rPos me)
 where
  me = findMe state self

moveToBaseStrategy::UArray Point Int->Board->Self->State->[Point]->Strategy
moveToBaseStrategy dists board self state bases =
 Strategy {strName = "Move to base " ++ show base,
           strGenerate = genMoveToBaseMoves base dDists,
           strCheck = genCheck,
           strHandlers = [attackHandler False 1, emergencyHandler, pickHandler 3, checkHandler 3]}
 where
  (_,base) = minimum $ map (\b -> (dists ! b,b)) bases
  dDists = distances board base
  genCheck b s st (Move _ (MMove dir)) =
    checkAt (rPos (findMe st s) `move` dir) `andCheck`
    checkBaseExists base

genMoveToFreePackageMoves::Int->Point->UArray Point Int->Board->Self->State->Maybe [Move]
genMoveToFreePackageMoves pack to dists board self state
 | pPos (findPackage state pack) /= Left to = Nothing
 | otherwise = Just $ map (\d -> Move 1 (MMove d)) $ rPath dists (rPos me)
 where
  me = findMe state self

findFreePackagesStrategy::UArray Point Int->Board->Self->State->FiniteMap Int Package->Strategy
findFreePackagesStrategy dists board self state packages =
 Strategy {strName = "Find free packages at " ++ show pos,
           strGenerate = genMoveToFreePackageMoves pack pos dDists,
           strCheck = genCheck,
           strHandlers = [attackHandler False 1, emergencyHandler, pickHandler 3, checkHandler 3]}
 where
  ttws = map (countTTW state dists) (keysFM packages)
  (_,pack) = minimum ttws
  pkg = findPackage state pack
  Left pos = pPos pkg
  dDists = distances board pos

  genCheck board self state (Move _ (MMove dir)) =
     checkPackageAt pack pos `andCheck`
     checkAt (rPos (findMe state self) `move` dir)

stayStrategy::Strategy
stayStrategy=
 Strategy {strName = "Do nothing",
           strGenerate = \_ _ _ -> Just $ cycle [Move 1 (MDrop [])],
           strCheck = \_ _ _ _ -> alwaysOK,
           strHandlers = []}

genHuntMoves::Int->Board->Self->State->Maybe [Move]
genHuntMoves i board self state
 | not (checkAlive i board self state) = Nothing
 | not (checkHoldsPackage i board self state) = Nothing
 | checkFreePackages board self state = Nothing
 | otherwise = Just $ map (\d->Move 1 (MMove d)) $ path dists (rPos he)
 where
  me = findMe state self
  he = findRobot state i
  dists = distances board (rPos me)

huntStrategy::UArray Point Int->Board->Self->State->FiniteMap Int Package->Strategy
huntStrategy dists board self state packages =
 Strategy {strName = "Hunt #" ++ show r,
           strGenerate = genHuntMoves r,
           strCheck = genCheck,
           strHandlers = [attackHandler True 5, emergencyHandler]}
 where
  robots = map (\(Right x)->x) $ map pPos $ eltsFM packages
  rdists = map (\i-> (dists ! (rPos $ findRobot state i), i)) robots
  (_,r) = minimum rdists

  genCheck b s st (Move _ (MMove dir)) =
    checkAlive r `andCheck`
    checkAt (rPos me `move` dir) `andCheck`
    checkHoldsPackage r `andCheck`
    (notCheck checkFreePackages)
   where
    me = findMe st s

findPackagesStrategy::UArray Point Int->Board->Self->State->Strategy
findPackagesStrategy dists board self state
 | not (isEmptyFM sure) = findFreePackagesStrategy dists board self state sure
 | not (isEmptySet bases) = moveToBaseStrategy dists board self state (setToList bases)
 | not (isEmptyFM free) = findFreePackagesStrategy dists board self state free
 | not (isEmptyFM carried) = huntStrategy dists board self state carried
 | otherwise = stayStrategy
 where
  packages = sPackages state
  bases = sBases state
  sure = filterFM (\_ -> testWeight self) $ filterFM (\_ -> testDestination board) $
         filterFM (\_ pack -> isLeft (pPos pack) && pSure pack) packages
  free = filterFM (\_ -> testWeight self) $ filterFM (\_ -> testDestination board) $
         filterFM (\_ pack -> isLeft (pPos pack) && not (pSure pack)) packages
  carried = filterFM (\_ -> testWeight self) $ filterFM (\_ -> testDestination board) $
            filterFM (\_ pack -> isRight (pPos pack)) packages
  
genPickMove::[Int]->Board->Self->State->Maybe [Move]
genPickMove l board self state
  | checkUnpickable l board self state = Nothing 
  | otherwise = Just [Move (if length l > 1 then 3 else 1) (MPick l)] 

gatherPackagesStrategy::UArray Point Int->Board->Self->State->FiniteMap Int Package->Strategy
gatherPackagesStrategy times board self state toPick =
 Strategy {strName = "Gather packages " ++ show cToPick,
           strGenerate = genPickMove cToPick,
           strCheck = \b s st (Move _ (MPick l)) -> checkUnpickable l,
           strHandlers = [attackHandler False 1, emergencyHandler]}
 where
  pDests = map (\l-> takeMost capaLeft $
                       reverse $ sort $
                       map (\(_,i,p)->(fromJust (pWeight p),i)) l) $
           groupBy (\(d1,_,_) (d2,_,_)->d1 == d2) $
           sortBy (\(d1,_,_) (d2,_,_)->compare d1 d2) $
           map getInfo $ fmToList toPick
  ttWs = map deliTTW pDests
  uList = sortBy (\(c1,_) (c2,_)->compare c1 c2) ttWs
  capaLeft = capa state self
  cToPick = map snd $ takeMost capaLeft $ concat $ map snd uList

  me = findMe state self
  myPacks = setToList (rPackages me)
  myW = sCapa self - capaLeft

  deliTTW l = (fst (deliveryTime (is++myPacks) (rPos me) (Just times) state) `div` (tW + myW),l)
   where
    weights = map fst l
    is = map snd l
    tW = sum weights
  
  getInfo (i,pack) = (fromJust (pDest pack),i,pack)

  takeMost _ [] = []
  takeMost capa ((w,i):rest)
   | w > capa = takeMost capa rest
   | otherwise = (w,i) : takeMost (capa - w) rest

genDeliverMoves::Int->Point->UArray Point Int->Board->Self->State->Maybe [Move]
genDeliverMoves what to dists board self state
 | not (checkHave what board self state) = Nothing
 | otherwise = Just $ (map (\d->Move 1 (MMove d)) $ rPath dists $ rPos me) ++ [dropDelivered]
 where
  me = findMe state self
  pack = findPackage state what
  myPackages = setToList $ rPackages me
  dropDelivered = Move 3 (MDrop toDrop)
  toDrop = filter (shouldDropAt to) myPackages 
  shouldDropAt pos i = (fromJust $ pDest $ findPackage state i) == pos

deliverPackagesStrategy::UArray Point Int->Board->Self->State->Strategy
deliverPackagesStrategy dDists board self state =
 Strategy {strName = "Deliver package " ++ show toDeliver ++ " to " ++ show pos,
           strGenerate = genDeliverMoves toDeliver pos dists,
           strCheck = genCheck,
           strHandlers = [attackHandler False 1, emergencyHandler, pickHandler 3, checkHandler 3]}
 where
  me = findMe state self
  have = setToList $ rPackages me
  toDeliver = snd $ deliveryTime have (rPos me) (Just dDists) state
  pToDeliver = findPackage state toDeliver
  pos =fromJust $ pDest pToDeliver
  dists = distances board pos

  genCheck b s st (Move _ (MMove dir)) = checkHave toDeliver `andCheck` checkAt (rPos me `move` dir)
   where
    me = findMe st s
  genCheck b s st (Move _ (MDrop list)) = checkDelivered list

deliveryTime::[Int]->Point->Maybe (UArray Point Int)->State->(Int,Int)
deliveryTime packs from dists state = (c,i)
 where
  pDests = map (\l->((\(d,_,_)->d) $ head l,
                     (\(_,i,_)->i) $ head l,
                     sum $ map (\(_,_,p)->fromJust (pWeight p)) l)) $
           groupBy (\(d1,_,_) (d2,_,_)->d1 == d2) $
           sortBy (\(d1,_,_) (d2,_,_)->compare d1 d2) $ map getInfo packs
  sDests = listToFM $ map (\(d,i,w)->(d,(i,w))) pDests
  bsDests = map checkThisStart $ fmToList sDests

  (c,(i,_)) = minimumBy (\(c1,_) (c2,_)->compare c1 c2) bsDests

  checkThisStart (dest,nfo) = (cDist from dest + c, nfo)
   where
    (c,_) = hDeliv dest (sDests `delFromFM` dest)
  
  hDeliv from dests
   | isEmptyFM dests = (0,undefined)
   | otherwise = (c+rCost,rDt)
   where
    (c,dest,rDt) = minimumBy (\(c1,_,_) (c2,_,_)->compare c1 c2) dCosts
    (rCost,_) = hDeliv dest (dests `delFromFM` dest)
    dCosts = map (\(d,dt)->(cost d,d,dt)) (fmToList dests)
    cost dest = cDist from dest

  getInfo i = (fromJust (pDest pack),i,pack)
   where
    pack = findPackage state i
  cDist p1 p2
   | p1==from, Just jDists <- dists = jDists ! p2
   | p2==from, Just jDists <- dists = jDists ! p1
   | otherwise = vic p1 p2

defaultStrategy::Board->Self->State->Strategy
defaultStrategy board self state
 | not (isEmptyFM pickablePackages) = gatherPackagesStrategy dists board self state pickablePackages
 | isEmptySet myPackages = findPackagesStrategy dists board self state
 | otherwise = deliverPackagesStrategy dists board self state
 where
  packages = sPackages state
  me = findMe state self
  myPackages = rPackages me
  dists = distances board (rPos me)
  capaLeft = capa state self
  packagesHere = filterFM (\_ pack -> pPos pack == Left (rPos me)) packages
  pickablePackages = filterFM (\_ -> testDestination board) $
                     filterFM (\_ pack -> fromJust (pWeight pack) <= capaLeft) packagesHere

